var assert = require("assert");
var Lib = require("./../ramda");

describe('or', function() {
    var or = Lib.or;

    it('should combine two boolean-returning functions into one', function() {
        var even = function(x) {return x % 2 === 0;};
        var gt10 = function(x) {return x > 10;};
        var f = or(even, gt10);
        assert.equal(f(8), true);
        assert.equal(f(13), true);
        assert.equal(f(7), false);
    });

    it('should accept functions that take multiple parameters', function() {
        var between = function(a, b, c) {return a < b && b < c;};
        var total20 = function(a, b, c) {return a + b + c === 20;};
        var f = or(between, total20);
        assert.equal(f(4, 5, 8), true);
        assert.equal(f(12, 2, 6), true);
        assert.equal(f(7, 5, 1), false);
    });

    it("does not evaluate the second expression if the first one is true", function() {
        var T = function() { return true; };
        var Z = function() { effect = "Z got evaluated"; };
        var effect = "not evaluated";
        or(T, Z);
        assert.equal(effect, "not evaluated");
    });
});

describe('and', function() {
    var and = Lib.and;

    it('should combine two boolean-returning functions into one', function() {
        var even = function(x) {return x % 2 === 0;};
        var gt10 = function(x) {return x > 10;};
        var f = and(even, gt10);
        assert.equal(f(8), false);
        assert.equal(f(13), false);
        assert.equal(f(14), true);
    });

    it('should accept functions that take multiple parameters', function() {
        var between = function(a, b, c) {return a < b && b < c;};
        var total20 = function(a, b, c) {return a + b + c === 20;};
        var f = and(between, total20);
        assert.equal(f(4, 5, 11), true);
        assert.equal(f(12, 2, 6), false);
        assert.equal(f(5, 6, 15), false);
    });

    it("does not evaluate the second expression if the first one is false", function() {
        var F = function() { return false; };
        var Z = function() { effect = "Z got evaluated"; };
        var effect = "not evaluated";
        and(F, Z);
        assert.equal(effect, "not evaluated");
    });
});

describe('not', function() {
    var not = Lib.not;

    it('should create boolean-returning function that reverses another', function() {
        var even = function(x) {return x % 2 === 0;};
        var f = not(even);
        assert.equal(f(8), false);
        assert.equal(f(13), true);
    });

    it('should accept a function that take multiple parameters', function() {
        var between = function(a, b, c) {return a < b && b < c;};
        var f = not(between);
        assert.equal(f(4, 5, 11), false);
        assert.equal(f(12, 2, 6), true);
    });
});

describe('allPredicates', function() {
    var allPredicates = Lib.allPredicates;
    var odd = function(n) {return !!(n % 2);};
    var lt20 = function(n) {return n < 20;};
    var gt5 = function(n) {return n > 5;};
    var plusEq = function(w, x, y, z) { return w + x  === y + z };

    it('should report whether all predicates are satisfied by a given value', function() {
        var ok = allPredicates([odd, lt20, gt5]);
        assert.equal(ok(7), true);
        assert.equal(ok(9), true);
        assert.equal(ok(10), false);
        assert.equal(ok(3), false);
        assert.equal(ok(21), false);
    });

    it('does not have to be curried', function() {
        assert.equal(allPredicates([odd, gt5], 3), false);
        assert.equal(allPredicates([odd, gt5], 7), true);
    });

    it('reports its arity as the longest predicate length', function() {
        assert.equal(allPredicates([odd, gt5, plusEq]).length, 4);
    });
});

describe('anyPredicates', function() {
    var anyPredicates = Lib.anyPredicates;
    var odd = function(n) {return !!(n % 2);};
    var gt20 = function(n) {return n > 20;};
    var lt5 = function(n) {return n < 5;};
    var plusEq = function(w, x, y, z) { return w + x  === y + z };

    it('should report whether any predicates are satisfied by a given value', function() {
        var ok = anyPredicates([odd, gt20, lt5]);
        assert.equal(ok(7), true);
        assert.equal(ok(9), true);
        assert.equal(ok(10), false);
        assert.equal(ok(18), false);
        assert.equal(ok(3), true);
        assert.equal(ok(22), true);
    });

    it('does not have to be curried', function() {
        assert.equal(anyPredicates([odd, lt5], 3), true);
        assert.equal(anyPredicates([odd, lt5], 22), false);
    });

    it('reports its arity as the longest predicate length', function() {
      assert.equal(anyPredicates([odd, lt5, plusEq]).length, 4);
    });
});
